import os

def create_key(template, outtype=('nii.gz','dicom'), annotation_classes=None):
    if template is None or not template:
        raise ValueError('Template must be a valid format string')
    return (template, outtype, annotation_classes)


def infotodict(seqinfo):
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    t1 = create_key('sub-{subject}/anat/sub-{subject}_T1w')
    t2 = create_key('sub-{subject}/anat/sub-{subject}_T2w')
    rest = create_key('sub-{subject}/func/sub-{subject}_task-rest_run-{item:02d}_bold')
    rest_sbref = create_key('sub-{subject}/func/sub-{subject}_task-rest_run-{item:02d}_sbref')
    face = create_key('sub-{subject}/func/sub-{subject}_task-face_run-{item:02d}_bold')
    face_sbref = create_key('sub-{subject}/func/sub-{subject}_task-face_run-{item:02d}_sbref')
    gamble = create_key('sub-{subject}/func/sub-{subject}_task-gambling_run-{item:02d}_bold')
    gamble_sbref = create_key('sub-{subject}/func/sub-{subject}_task-gambling_run-{item:02d}_sbref')
    conflict = create_key('sub-{subject}/func/sub-{subject}_task-conflict_run-{item:02d}_bold')
    conflict_sbref = create_key('sub-{subject}/func/sub-{subject}_task-conflict_run-{item:02d}_sbref')
    dwi = create_key('sub-{subject}/dwi/sub-{subject}_run-{item:02d}_dwi')
    dwi_sbref = create_key('sub-{subject}/dwi/sub-{subject}_run-{item:02d}_sbref')
    fmap = create_key('sub-{subject}/fmap/sub-{subject}_dir-{dir}_run-{item:02d}_epi')

    info = {t1:[], t2:[],
            rest:[], face:[], gamble:[], conflict:[], dwi:[],
            rest_sbref:[], face_sbref:[], gamble_sbref:[], conflict_sbref:[], dwi_sbref:[],
            fmap:[]}

    for idx, s in enumerate(seqinfo):
        # T1 and T2 scans
        if (s.dim3 == 208) and (s.dim4 == 1) and ('T1w' in s.protocol_name):
            info[t1] = [s.series_id]
        if (s.dim3 == 208) and ('T2w' in s.protocol_name):
            info[t2] = [s.series_id]
        # diffusion scans
        if ('dMRI_dir9' in s.protocol_name):
            key = None
            if (s.dim4 >= 99):
                key = dwi
            elif (s.dim4 == 1) and ('SBRef' in s.series_description):
                key = dwi_sbref
            if key:
                info[key].append({'item': s.series_id})
        # functional scans
        if ('fMRI' in s.protocol_name):
            tasktype = s.protocol_name.split('fMRI')[1].split('_')[1]
            key = None
            if (s.dim4 in [420, 215, 338, 280]):
                if 'rest' in tasktype: key = rest
                if 'face' in tasktype: key = face
                if 'conflict' in tasktype: key = conflict
                if 'gambling' in tasktype: key = gamble
            if (s.dim4 == 1) and ('SBRef' in s.series_description):
                if 'rest' in tasktype: key = rest_sbref
                if 'face' in tasktype: key = face_sbref
                if 'conflict' in tasktype: key = conflict_sbref
                if 'gambling' in tasktype: key = gamble_sbref
            if key:
                info[key].append({'item': s.series_id})
        if (s.dim4 == 3) and ('SpinEchoFieldMap' in s.protocol_name):
            dirtype = s.protocol_name.split('_')[-1]
            info[fmap].append({'item': s.series_id, 'dir': dirtype})

    # You can even put checks in place for your protocol
    msg = []
    if len(info[t1]) != 1: msg.append('Missing correct number of t1 runs')
    if len(info[t2]) != 1: msg.append('Missing correct number of t2 runs')
    if len(info[dwi]) != 4: msg.append('Missing correct number of dwi runs')
    if len(info[rest]) != 4: msg.append('Missing correct number of resting runs')
    if len(info[face]) != 2: msg.append('Missing correct number of faceMatching runs')
    if len(info[conflict]) != 4: msg.append('Missing correct number of conflict runs')
    if len(info[gamble]) != 2: msg.append('Missing correct number of gamble runs')
    if msg:
        raise ValueError('\n'.join(msg))
    return info



import os.path as op
import heudiconv.heuristics

HEURISTICS_PATH = op.join(heudiconv.heuristics.__path__[0])
TESTS_DATA_PATH = op.join(op.dirname(__file__), 'data')


def gen_heudiconv_args(datadir, outdir, subject, heuristic_file, xargs=None):
    heuristic = op.realpath(op.join(HEURISTICS_PATH, heuristic_file))
    args = ["-d", op.join(datadir, 'sourcedata/{subject}/*/*/*.tgz'),
            "-c", "dcm2niix",
            "-o", outdir,
            "-s", subject,
            "-f", heuristic,
            "--bids",]
    if xargs:
        args += xargs

    return args


def fetch_data(tmpdir, subject):
    """Fetches some test dicoms using datalad"""
    from datalad import api
    targetdir = op.join(tmpdir, 'QA')
    api.install(path=targetdir, source='http://datasets-tests.datalad.org/dbic/QA')
    api.get('{}/sourcedata/{}'.format(targetdir, subject))
    return targetdir



import os


def create_key(template, outtype=('nii.gz',), annotation_classes=None):
    if template is None or not template:
        raise ValueError('Template must be a valid format string')
    return template, outtype, annotation_classes


def infotodict(seqinfo):
    """Heuristic evaluator for determining which runs belong where
    
    allowed template fields - follow python string module: 
    
    item: index within category 
    subject: participant id 
    seqitem: run number during scanning
    subindex: sub index within group
    """
    
    rs = create_key('rsfmri/rest_run{item:03d}/rest', outtype=('dicom', 'nii.gz'))
    boldt1 = create_key('BOLD/task001_run{item:03d}/bold')
    boldt2 = create_key('BOLD/task002_run{item:03d}/bold')
    boldt3 = create_key('BOLD/task003_run{item:03d}/bold')
    boldt4 = create_key('BOLD/task004_run{item:03d}/bold')
    boldt5 = create_key('BOLD/task005_run{item:03d}/bold')
    boldt6 = create_key('BOLD/task006_run{item:03d}/bold')
    boldt7 = create_key('BOLD/task007_run{item:03d}/bold')
    boldt8 = create_key('BOLD/task008_run{item:03d}/bold')
    fm1 = create_key('fieldmap/fm1_{item:03d}')
    fm2 = create_key('fieldmap/fm2_{item:03d}')
    fmrest = create_key('fieldmap/fmrest_{item:03d}')
    dwi = create_key('dmri/dwi_{item:03d}', outtype=('dicom', 'nii.gz'))
    t1 = create_key('anatomy/T1_{item:03d}')
    asl = create_key('rsfmri/asl_run{item:03d}/asl')
    aslcal = create_key('rsfmri/asl_run{item:03d}/cal_{subindex:03d}')
    info = {rs: [], boldt1: [], boldt2: [], boldt3: [], boldt4: [], 
            boldt5: [], boldt6: [], boldt7: [], boldt8: [],
            fm1: [], fm2: [], fmrest: [], dwi: [], t1: [], 
            asl: [], aslcal: [[]]}
    last_run = len(seqinfo)
    for s in seqinfo:
        x, y, sl, nt = (s[6], s[7], s[8], s[9])
        if (sl == 176) and (nt == 1) and ('MPRAGE' in s[12]):
            info[t1] = [s[2]]
        elif (nt > 60) and ('ge_func_2x2x2_Resting' in s[12]):
            if not s[13]:
                info[rs].append(int(s[2]))
        elif (nt == 156) and ('ge_functionals_128_PACE_ACPC-30' in s[12]) and s[2] < last_run:
            if not s[13]:
                info[boldt1].append(s[2])
                last_run = s[2]
        elif (nt == 155) and ('ge_functionals_128_PACE_ACPC-30' in s[12]):
            if not s[13]:
                info[boldt2].append(s[2])
        elif (nt == 222) and ('ge_functionals_128_PACE_ACPC-30' in s[12]):
            if not s[13]:
                info[boldt3].append(s[2])
        elif (nt == 114) and ('ge_functionals_128_PACE_ACPC-30' in s[12]):
            if not s[13]:
                info[boldt4].append(s[2])
        elif (nt == 156) and ('ge_functionals_128_PACE_ACPC-30' in s[12]):
            if not s[13] and (s[2] > last_run):
                info[boldt5].append(s[2])
        elif (nt == 324) and ('ge_func_3.1x3.1x4_PACE' in s[12]):
            if not s[13]:
                info[boldt6].append(s[2])
        elif (nt == 250) and ('ge_func_3.1x3.1x4_PACE' in s[12]):
            if not s[13]:
                info[boldt7].append(s[2])
        elif (nt == 136) and ('ge_func_3.1x3.1x4_PACE' in s[12]):
            if not s[13]:
                info[boldt8].append(s[2])
        elif (nt == 101) and ('ep2d_pasl_FairQuipssII' in s[12]):
            if not s[13]:
                info[asl].append(s[2])
        elif (nt == 1) and ('ep2d_pasl_FairQuipssII' in s[12]):
            info[aslcal][0].append(s[2])
        elif (sl > 1) and (nt == 70) and ('DIFFUSION' in s[12]):
            info[dwi].append(s[2])
        elif ('field_mapping_128' in s[12]):
            info[fm1].append(s[2])
        elif ('field_mapping_3.1' in s[12]):
            info[fm2].append(s[2])
        elif ('field_mapping_Resting' in s[12]):
            info[fmrest].append(s[2])
        else:
            pass
    return info



"""Testing conversion with conversion saved on datalad"""
import json
from glob import glob

import pytest

have_datalad = True
try:
    from datalad import api # to pull and grab data
    from datalad.support.exceptions import IncompleteResultsError
except ImportError:
    have_datalad = False

import heudiconv
from heudiconv.cli.run import main as runner
# testing utilities
from .utils import fetch_data, gen_heudiconv_args


@pytest.mark.parametrize('subject', ['sub-sid000143'])
@pytest.mark.parametrize('heuristic', ['reproin.py'])
@pytest.mark.skipif(not have_datalad, reason="no datalad")
def test_conversion(tmpdir, subject, heuristic):
    tmpdir.chdir()
    try:
        datadir = fetch_data(tmpdir.strpath, subject)
    except IncompleteResultsError as exc:
        pytest.skip("Failed to fetch test data: %s" % str(exc))
    outdir = tmpdir.mkdir('out').strpath

    args = gen_heudiconv_args(datadir, outdir, subject, heuristic)
    runner(args) # run conversion

    # verify functionals were converted
    assert glob('{}/{}/func/*'.format(outdir, subject)) == \
           glob('{}/{}/func/*'.format(datadir, subject))

    # compare some json metadata
    json_ = '{}/task-rest_acq-24mm64sl1000tr32te600dyn_bold.json'.format
    orig, conv = (json.load(open(json_(datadir))),
                  json.load(open(json_(outdir))))
    keys = ['EchoTime', 'MagneticFieldStrength', 'Manufacturer', 'SliceTiming']
    for key in keys:
        assert orig[key] == conv[key]



import os


def create_key(template, outtype=('nii.gz',), annotation_classes=None):
    if template is None or not template:
        raise ValueError('Template must be a valid format string')
    return template, outtype, annotation_classes


def infotodict(seqinfo):
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    session: scan index for longitudinal acq
    """
    # for this example, we want to include copies of the DICOMs just for our T1
    # and functional scans
    outdicom = ('dicom', 'nii.gz')

    t1 = create_key('{bids_subject_session_dir}/anat/{bids_subject_session_prefix}_T1w', outtype=outdicom)
    t2 = create_key('{bids_subject_session_dir}/anat/{bids_subject_session_prefix}_T2w')
    dwi_ap = create_key('{bids_subject_session_dir}/dwi/{bids_subject_session_prefix}_dir-AP_dwi')
    dwi_pa = create_key('{bids_subject_session_dir}/dwi/{bids_subject_session_prefix}_dir-PA_dwi')
    rs = create_key('{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-rest_run-{item:02d}_bold', outtype=outdicom)
    boldt1 = create_key('{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-bird1back_run-{item:02d}_bold', outtype=outdicom)
    boldt2 = create_key('{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-letter1back_run-{item:02d}_bold', outtype=outdicom)
    boldt3 = create_key('{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-letter2back_run-{item:02d}_bold', outtype=outdicom)

    info = {t1: [], t2:[], dwi_ap:[], dwi_pa:[], rs:[],
            boldt1:[], boldt2:[], boldt3:[],}
    last_run = len(seqinfo)
    for s in seqinfo:
        if (s.dim3 == 176 or s.dim3 == 352) and (s.dim4 == 1) and ('MEMPRAGE' in s.protocol_name):
            info[t1] = [s.series_id]
        elif (s.dim4 == 1) and ('MEMPRAGE' in s.protocol_name):
            info[t1] = [s.series_id]
        elif (s.dim3 == 176 or s.dim3 == 352) and (s.dim4 == 1) and ('T2_SPACE' in s.protocol_name):
            info[t2] = [s.series_id]
        elif (s.dim4 >= 70) and ('DIFFUSION_HighRes_AP' in s.protocol_name):
            info[dwi_ap].append([s.series_id])
        elif ('DIFFUSION_HighRes_PA' in s.protocol_name):
            info[dwi_pa].append([s.series_id])
        elif (s.dim4 == 144) and ('resting' in s.protocol_name):
            if not s.is_motion_corrected:
                info[rs].append([(s.series_id)])
        elif (s.dim4 == 183 or s.dim4 == 366) and ('localizer' in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt1].append([s.series_id])
        elif (s.dim4 == 227 or s.dim4 == 454) and ('transfer1' in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt2].append([s.series_id])
        elif (s.dim4 == 227 or s.dim4 == 454) and ('transfer2' in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt3].append([s.series_id])
    return info



# TODO: break this up by modules

from heudiconv.cli.run import main as runner
from heudiconv import __version__
from heudiconv.utils import (create_file_if_missing,
                             set_readonly,
                             is_readonly)
from heudiconv.bids import (populate_bids_templates,
                            add_participant_record,
                            get_formatted_scans_key_row,
                            add_rows_to_scans_keys_file,
                            find_subj_ses)
from heudiconv.external.dlad import MIN_VERSION, add_to_datalad

from .utils import TESTS_DATA_PATH
import csv
import os
import pytest
import sys

from mock import patch
from os.path import join as opj
from six.moves import StringIO
import stat


@patch('sys.stdout', new_callable=StringIO)
def test_main_help(stdout):
    with pytest.raises(SystemExit):
        runner(['--help'])
    assert stdout.getvalue().startswith("usage: ")


@patch('sys.stderr' if sys.version_info[:2] <= (3, 3) else
       'sys.stdout', new_callable=StringIO)
def test_main_version(std):
    with pytest.raises(SystemExit):
        runner(['--version'])
    assert std.getvalue().rstrip() == __version__


def test_create_file_if_missing(tmpdir):
    tf = tmpdir.join("README.txt")
    assert not tf.exists()
    create_file_if_missing(str(tf), "content")
    assert tf.exists()
    assert tf.read() == "content"
    create_file_if_missing(str(tf), "content2")
    # nothing gets changed
    assert tf.read() == "content"


def test_populate_bids_templates(tmpdir):
    populate_bids_templates(
        str(tmpdir),
        defaults={'Acknowledgements': 'something'})
    for f in "README", "dataset_description.json", "CHANGES":
        # Just test that we have created them and they all have stuff TODO
        assert "TODO" in tmpdir.join(f).read()
    description_file = tmpdir.join('dataset_description.json')
    assert "something" in description_file.read()

    # it should also be available as a command
    os.unlink(str(description_file))
    runner([
        '--command', 'populate-templates', '-f', 'convertall',
        '--files', str(tmpdir)
    ])
    assert "something" not in description_file.read()
    assert "TODO" in description_file.read()


def test_add_participant_record(tmpdir):
    tf = tmpdir.join('participants.tsv')
    assert not tf.exists()
    add_participant_record(str(tmpdir), "sub01", "023Y", "M")
    # should create the file and place corrected record
    sub01 = tf.read()
    assert sub01 == """\
participant_id	age	sex	group
sub-sub01	23	M	control
"""
    add_participant_record(str(tmpdir), "sub01", "023Y", "F")
    assert tf.read() == sub01  # nothing was added even though differs in values
    add_participant_record(str(tmpdir), "sub02", "2", "F")
    assert tf.read() == """\
participant_id	age	sex	group
sub-sub01	23	M	control
sub-sub02	2	F	control
"""


def test_prepare_for_datalad(tmpdir):
    pytest.importorskip("datalad", minversion=MIN_VERSION)
    studydir = tmpdir.join("PI").join("study")
    studydir_ = str(studydir)
    os.makedirs(studydir_)
    populate_bids_templates(studydir_)

    add_to_datalad(str(tmpdir), studydir_, None, False)

    from datalad.api import Dataset
    superds = Dataset(str(tmpdir))

    assert superds.is_installed()
    assert not superds.repo.dirty
    subdss = superds.subdatasets(recursive=True, result_xfm='relpaths')
    for ds_path in sorted(subdss):
        ds = Dataset(opj(superds.path, ds_path))
        assert ds.is_installed()
        assert not ds.repo.dirty

    # the last one should have been the study
    target_files = {
        '.gitattributes',
        '.datalad/config', '.datalad/.gitattributes',
        'dataset_description.json',
        'CHANGES', 'README'}
    assert set(ds.repo.get_indexed_files()) == target_files
    # and all are under git
    for f in target_files:
        assert not ds.repo.is_under_annex(f)
    assert not ds.repo.is_under_annex('.gitattributes')

    # Above call to add_to_datalad does not create .heudiconv subds since
    # directory does not exist (yet).
    # Let's first check that it is safe to call it again
    add_to_datalad(str(tmpdir), studydir_, None, False)
    assert not ds.repo.dirty

    old_hexsha = ds.repo.get_hexsha()
    # Now let's check that if we had previously converted data so that
    # .heudiconv was not a submodule, we still would not fail
    dsh_path = os.path.join(ds.path, '.heudiconv')
    dummy_path = os.path.join(dsh_path, 'dummy.nii.gz')

    create_file_if_missing(dummy_path, '')
    ds.add(dummy_path, message="added a dummy file")
    # next call must not fail, should just issue a warning
    add_to_datalad(str(tmpdir), studydir_, None, False)
    ds.repo.is_under_annex(dummy_path)
    assert not ds.repo.dirty
    assert '.heudiconv/dummy.nii.gz' in ds.repo.get_files()

    # Let's now roll back and make it a proper submodule
    ds.repo._git_custom_command([], ['git', 'reset', '--hard', old_hexsha])
    # now we do not add dummy to git
    create_file_if_missing(dummy_path, '')
    add_to_datalad(str(tmpdir), studydir_, None, False)
    assert '.heudiconv' in ds.subdatasets(result_xfm='relpaths')
    assert not ds.repo.dirty
    assert '.heudiconv/dummy.nii.gz' not in ds.repo.get_files()


def test_get_formatted_scans_key_row():
    item = [
         ('%s/01-fmap_acq-3mm/1.3.12.2.1107.5.2.43.66112.2016101409263663466202201.dcm'
          % TESTS_DATA_PATH,
         ('nii.gz', 'dicom'),
         ['%s/01-fmap_acq-3mm/1.3.12.2.1107.5.2.43.66112.2016101409263663466202201.dcm'
          % TESTS_DATA_PATH])
    ]
    outname_bids_file = '/a/path/Halchenko/Yarik/950_bids_test4/sub-phantom1sid1/fmap/sub-phantom1sid1_acq-3mm_phasediff.json'

    row = get_formatted_scans_key_row(item)
    assert len(row) == 3
    assert row[0] == '2016-10-14T09:26:36'
    assert row[1] == 'n/a'
    randstr1 = row[2]
    row = get_formatted_scans_key_row(item)
    randstr2 = row[2]
    assert(randstr1 != randstr2)


# TODO: finish this
def test_add_rows_to_scans_keys_file(tmpdir):
    fn = opj(tmpdir.strpath, 'file.tsv')
    rows = {
        'my_file.nii.gz': ['2016adsfasd', '', 'fasadfasdf'],
        'another_file.nii.gz': ['2018xxxxx', '', 'fasadfasdf']
    }
    add_rows_to_scans_keys_file(fn, rows)

    def _check_rows(fn, rows):
        with open(fn, 'r') as csvfile:
            reader = csv.reader(csvfile, delimiter='\t')
            rows_loaded = []
            for row in reader:
                rows_loaded.append(row)
        for i, row_ in enumerate(rows_loaded):
            if i == 0:
                assert(row_ == ['filename', 'acq_time', 'operator', 'randstr'])
            else:
                assert(rows[row_[0]] == row_[1:])
        # dates, filename should be sorted (date "first", filename "second")
        dates = [(r[1], r[0]) for r in rows_loaded[1:]]
        assert dates == sorted(dates)

    _check_rows(fn, rows)
    # add a new one
    extra_rows = {
        'a_new_file.nii.gz': ['2016adsfasd23', '', 'fasadfasdf'],
        'my_file.nii.gz': ['2016adsfasd', '', 'fasadfasdf'],
        'another_file.nii.gz': ['2018xxxxx', '', 'fasadfasdf']
    }
    add_rows_to_scans_keys_file(fn, extra_rows)
    _check_rows(fn, extra_rows)


def test__find_subj_ses():
    assert find_subj_ses(
        '950_bids_test4/sub-phantom1sid1/fmap/'
        'sub-phantom1sid1_acq-3mm_phasediff.json') == ('phantom1sid1', None)
    assert find_subj_ses(
        'sub-s1/ses-s1/fmap/sub-s1_ses-s1_acq-3mm_phasediff.json') == ('s1',
                                                                       's1')
    assert find_subj_ses(
        'sub-s1/ses-s1/fmap/sub-s1_ses-s1_acq-3mm_phasediff.json') == ('s1',
                                                                       's1')
    assert find_subj_ses(
        'fmap/sub-01-fmap_acq-3mm_acq-3mm_phasediff.nii.gz') == ('01', None)


def test_make_readonly(tmpdir):
    # we could test it all without torturing a poor file, but for going all
    # the way, let's do it on a file
    path = tmpdir.join('f')
    pathname = str(path)
    with open(pathname, 'w'):
        pass

    for orig, ro, rw in [
        (0o600, 0o400, 0o600),  # fully returned
        (0o624, 0o404, 0o606),  # it will not get write bit where it is not readable
        (0o1777, 0o1555, 0o1777),  # and other bits should be preserved
    ]:
        os.chmod(pathname, orig)
        assert not is_readonly(pathname)
        assert set_readonly(pathname) == ro
        assert is_readonly(pathname)
        assert stat.S_IMODE(os.lstat(pathname).st_mode) == ro
        # and it should go back if we set it back to non-read_only
        assert set_readonly(pathname, read_only=False) == rw
        assert not is_readonly(pathname)



import os
import os.path as op
from argparse import ArgumentParser
import sys

from .. import __version__, __packagename__
from ..parser import get_study_sessions
from ..utils import load_heuristic, anonymize_sid, treat_infofile, SeqInfo
from ..convert import prep_conversion
from ..bids import populate_bids_templates, tuneup_bids_json_files
from ..queue import queue_conversion

import inspect
import logging
lgr = logging.getLogger(__name__)

INIT_MSG = "Running {packname} version {version}".format


def is_interactive():
   """Return True if all in/outs are tty"""
   # TODO: check on windows if hasattr check would work correctly and add value:
   return sys.stdin.isatty() and sys.stdout.isatty() and sys.stderr.isatty()


def setup_exceptionhook():
    """
    Overloads default sys.excepthook with our exceptionhook handler.

    If interactive, our exceptionhook handler will invoke pdb.post_mortem;
    if not interactive, then invokes default handler.
    """
    def _pdb_excepthook(type, value, tb):
        if is_interactive():
            import traceback
            import pdb
            traceback.print_exception(type, value, tb)
            # print()
            pdb.post_mortem(tb)
        else:
            lgr.warning(
              "We cannot setup exception hook since not in interactive mode")

    sys.excepthook = _pdb_excepthook


def process_extra_commands(outdir, args):
    """
    Perform custom command instead of regular operations. Supported commands:
    ['treat-json', 'ls', 'populate-templates']

    Parameters
    ----------
    outdir : String
        Output directory
    args : Namespace
        arguments
    """
    if args.command == 'treat-jsons':
        for f in args.files:
            treat_infofile(f)
    elif args.command == 'ls':
        heuristic = load_heuristic(args.heuristic)
        heuristic_ls = getattr(heuristic, 'ls', None)
        for f in args.files:
            study_sessions = get_study_sessions(
                args.dicom_dir_template, [f], heuristic, outdir,
                args.session, args.subjs, grouping=args.grouping)
            print(f)
            for study_session, sequences in study_sessions.items():
                suf = ''
                if heuristic_ls:
                    suf += heuristic_ls(study_session, sequences)
                print(
                    "\t%s %d sequences%s"
                    % (str(study_session), len(sequences), suf)
                )
    elif args.command == 'populate-templates':
        heuristic = load_heuristic(args.heuristic)
        for f in args.files:
            populate_bids_templates(f, getattr(heuristic, 'DEFAULT_FIELDS', {}))
    elif args.command == 'sanitize-jsons':
        tuneup_bids_json_files(args.files)
    elif args.command == 'heuristics':
        from ..utils import get_known_heuristics_with_descriptions
        for name_desc in get_known_heuristics_with_descriptions().items():
            print("- %s: %s" % name_desc)
    elif args.command == 'heuristic-info':
        from ..utils import get_heuristic_description, get_known_heuristic_names
        if not args.heuristic:
            raise ValueError("Specify heuristic using -f. Known are: %s"
                             % ', '.join(get_known_heuristic_names()))
        print(get_heuristic_description(args.heuristic, full=True))
    else:
        raise ValueError("Unknown command %s", args.command)
    return


def main(argv=None):
    parser = get_parser()
    args = parser.parse_args(argv)
    # exit if nothing to be done
    if not args.files and not args.dicom_dir_template and not args.command:
        lgr.warning("Nothing to be done - displaying usage help")
        parser.print_help()
        sys.exit(1)
    # To be done asap so anything random is deterministic
    if args.random_seed is not None:
        import random
        random.seed(args.random_seed)
        import numpy
        numpy.random.seed(args.random_seed)
    if args.debug:
        lgr.setLevel(logging.DEBUG)
    # Should be possible but only with a single subject -- will be used to
    # override subject deduced from the DICOMs
    if args.files and args.subjs and len(args.subjs) > 1:
        raise ValueError(
            "Unable to processes multiple `--subjects` with files"
        )

    if args.debug:
        setup_exceptionhook()

    process_args(args)


def get_parser():
    docstr = ("""Example:
             heudiconv -d rawdata/{subject} -o . -f heuristic.py -s s1 s2 s3""")
    parser = ArgumentParser(description=docstr)
    parser.add_argument('--version', action='version', version=__version__)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-d', '--dicom_dir_template', dest='dicom_dir_template',
                       help='location of dicomdir that can be indexed with '
                       'subject id {subject} and session {session}. Tarballs '
                       '(can be compressed) are supported in addition to '
                       'directory. All matching tarballs for a subject are '
                       'extracted and their content processed in a single pass')
    group.add_argument('--files', nargs='*',
                       help='Files (tarballs, dicoms) or directories '
                       'containing files to process. Cannot be provided if '
                       'using --dicom_dir_template or --subjects')
    parser.add_argument('-s', '--subjects', dest='subjs', type=str, nargs='*',
                        help='list of subjects - required for dicom template. '
                        'If not provided, DICOMS would first be "sorted" and '
                        'subject IDs deduced by the heuristic')
    parser.add_argument('-c', '--converter',
                        default='dcm2niix',
                        choices=('dcm2niix', 'none'),
                        help='tool to use for DICOM conversion. Setting to '
                        '"none" disables the actual conversion step -- useful'
                        'for testing heuristics.')
    parser.add_argument('-o', '--outdir', default=os.getcwd(),
                        help='output directory for conversion setup (for '
                        'further customization and future reference. This '
                        'directory will refer to non-anonymized subject IDs')
    parser.add_argument('-l', '--locator', default=None,
                        help='study path under outdir.  If provided, '
                        'it overloads the value provided by the heuristic. '
                        'If --datalad is enabled, every directory within '
                        'locator becomes a super-dataset thus establishing a '
                        'hierarchy. Setting to "unknown" will skip that dataset')
    parser.add_argument('-a', '--conv-outdir', default=None,
                        help='output directory for converted files. By default '
                        'this is identical to --outdir. This option is most '
                        'useful in combination with --anon-cmd')
    parser.add_argument('--anon-cmd', default=None,
                        help='command to run to convert subject IDs used for '
                        'DICOMs to anonymized IDs. Such command must take a '
                        'single argument and return a single anonymized ID. '
                        'Also see --conv-outdir')
    parser.add_argument('-f', '--heuristic', dest='heuristic',
                        # some commands might not need heuristic
                        # required=True,
                        help='Name of a known heuristic or path to the Python'
                             'script containing heuristic')
    parser.add_argument('-p', '--with-prov', action='store_true',
                        help='Store additional provenance information. '
                        'Requires python-rdflib.')
    parser.add_argument('-ss', '--ses', dest='session', default=None,
                        help='session for longitudinal study_sessions, default '
                        'is none')
    parser.add_argument('-b', '--bids', action='store_true',
                        help='flag for output into BIDS structure')
    parser.add_argument('--overwrite', action='store_true', default=False,
                        help='flag to allow overwriting existing converted files')
    parser.add_argument('--datalad', action='store_true',
                        help='Store the entire collection as DataLad '
                        'dataset(s). Small files will be committed directly to '
                        'git, while large to annex. New version (6) of annex '
                        'repositories will be used in a "thin" mode so it '
                        'would look to mortals as just any other regular '
                        'directory (i.e. no symlinks to under .git/annex).  '
                        'For now just for BIDS mode.')
    parser.add_argument('--dbg', action='store_true', dest='debug',
                        help='Do not catch exceptions and show exception '
                        'traceback')
    parser.add_argument('--command',
                        choices=(
                            'heuristics', 'heuristic-info',
                            'ls', 'populate-templates',
                            'sanitize-jsons', 'treat-jsons',
                        ),
                        help='custom actions to be performed on provided '
                        'files instead of regular operation.')
    parser.add_argument('-g', '--grouping', default='studyUID',
                        choices=('studyUID', 'accession_number'),
                        help='How to group dicoms (default: by studyUID)')
    parser.add_argument('--minmeta', action='store_true',
                        help='Exclude dcmstack meta information in sidecar '
                        'jsons')
    parser.add_argument('--random-seed', type=int, default=None,
                        help='Random seed to initialize RNG')
    submission = parser.add_argument_group('Conversion submission options')
    submission.add_argument('-q', '--queue', default=None,
                            help='select batch system to submit jobs to instead'
                                 ' of running the conversion serially')
    submission.add_argument('--sbargs', dest='sbatch_args', default=None,
                            help='Additional sbatch arguments if running with '
                                 'queue arg')
    return parser


def process_args(args):
    """Given a structure of arguments from the parser perform computation"""

    # Deal with provided files or templates
    # pre-process provided list of files and possibly sort into groups/sessions
    # Group files per each study/sid/session

    outdir = op.abspath(args.outdir)

    if args.command:
        process_extra_commands(outdir, args)
        return

    lgr.info(INIT_MSG(packname=__packagename__,
                      version=__version__))


    #
    # Load heuristic -- better do it asap to make sure it loads correctly
    #
    if not args.heuristic:
        raise RuntimeError("No heuristic specified - add to arguments and rerun")

    heuristic = load_heuristic(args.heuristic)

    study_sessions = get_study_sessions(args.dicom_dir_template, args.files,
                                        heuristic, outdir, args.session,
                                        args.subjs, grouping=args.grouping)

    # extract tarballs, and replace their entries with expanded lists of files
    # TODO: we might need to sort so sessions are ordered???
    lgr.info("Need to process %d study sessions", len(study_sessions))

    # processed_studydirs = set()

    for (locator, session, sid), files_or_seqinfo in study_sessions.items():

        # Allow for session to be overloaded from command line
        if args.session is not None:
            session = args.session
        if args.locator is not None:
            locator = args.locator
        if not len(files_or_seqinfo):
            raise ValueError("nothing to process?")
        # that is how life is ATM :-/ since we don't do sorting if subj
        # template is provided
        if isinstance(files_or_seqinfo, dict):
            assert(isinstance(list(files_or_seqinfo.keys())[0], SeqInfo))
            dicoms = None
            seqinfo = files_or_seqinfo
        else:
            dicoms = files_or_seqinfo
            seqinfo = None

        if locator == 'unknown':
            lgr.warning("Skipping unknown locator dataset")
            continue

        if args.queue:
            if seqinfo and not dicoms:
                # flatten them all and provide into batching, which again
                # would group them... heh
                dicoms = sum(seqinfo.values(), [])
                raise NotImplementedError(
                    "we already grouped them so need to add a switch to avoid "
                    "any grouping, so no outdir prefix doubled etc")

            progname = op.abspath(inspect.getfile(inspect.currentframe()))

            queue_conversion(progname,
                             args.queue,
                             study_outdir,
                             heuristic.filename,
                             dicoms,
                             sid,
                             args.anon_cmd,
                             args.converter,
                             session,
                             args.with_prov,
                             args.bids)
            continue

        anon_sid = anonymize_sid(sid, args.anon_cmd) if args.anon_cmd else None
        if args.anon_cmd:
            lgr.info('Anonymized {} to {}'.format(sid, anon_sid))

        study_outdir = op.join(outdir, locator or '')
        anon_outdir = args.conv_outdir or outdir
        anon_study_outdir = op.join(anon_outdir, locator or '')

        # TODO: --datalad  cmdline option, which would take care about initiating
        # the outdir -> study_outdir datasets if not yet there
        if args.datalad:
            from ..external.dlad import prepare_datalad
            dlad_sid = sid if not anon_sid else anon_sid
            dl_msg = prepare_datalad(anon_study_outdir, anon_outdir, dlad_sid,
                                     session, seqinfo, dicoms, args.bids)

        lgr.info("PROCESSING STARTS: {0}".format(
            str(dict(subject=sid, outdir=study_outdir, session=session))))

        prep_conversion(sid,
                        dicoms,
                        study_outdir,
                        heuristic,
                        converter=args.converter,
                        anon_sid=anon_sid,
                        anon_outdir=anon_study_outdir,
                        with_prov=args.with_prov,
                        ses=session,
                        bids=args.bids,
                        seqinfo=seqinfo,
                        min_meta=args.minmeta,
                        overwrite=args.overwrite,)

        lgr.info("PROCESSING DONE: {0}".format(
            str(dict(subject=sid, outdir=study_outdir, session=session))))

        if args.datalad:
            from ..external.dlad import add_to_datalad
            msg = "Converted subject %s" % dl_msg
            # TODO:  whenever propagate to supers work -- do just
            # ds.save(msg=msg)
            #  also in batch mode might fail since we have no locking ATM
            #  and theoretically no need actually to save entire study
            #  we just need that
            add_to_datalad(outdir, study_outdir, msg, args.bids)

    # if args.bids:
    #     # Let's populate BIDS templates for folks to take care about
    #     for study_outdir in processed_studydirs:
    #         populate_bids_templates(study_outdir)
    #
    # TODO: record_collection of the sid/session although that information
    # is pretty much present in .heudiconv/SUBJECT/info so we could just poke there


if __name__ == "__main__":
    main()



import os
import os.path as op

import logging

lgr = logging.getLogger(__name__)

# start with SLURM but extend past that #TODO
def queue_conversion(progname, queue, outdir, heuristic, dicoms, sid,
                     anon_cmd, converter, session,with_prov, bids):

        # Rework this...
        convertcmd = ' '.join(['python', progname,
                               '-o', outdir,
                               '-f', heuristic,
                               '-s', sid,
                               '--anon-cmd', anon_cmd,
                               '-c', converter])
        if session:
            convertcmd += " --ses '%s'" % session
        if with_prov:
            convertcmd += " --with-prov"
        if bids:
            convertcmd += " --bids"
        if dicoms:
            convertcmd += " --files"
            convertcmd += [" '%s'" % f for f in dicoms]

        script_file = 'dicom-%s.sh' % sid
        with open(script_file, 'wt') as fp:
            fp.writelines(['#!/bin/bash\n', convertcmd])
        outcmd = 'sbatch -J dicom-%s -p %s -N1 -c2 --mem=20G %s' \
                 % (sid, queue, script_file)

        os.system(outcmd)



__version__ = "0.5.1"
__author__ = "HeuDiConv team and contributors"
__url__ = "https://github.com/nipy/heudiconv"
__packagename__ = 'heudiconv'
__description__ = "Heuristic DICOM Converter"
__license__ = "Apache 2.0"
__longdesc__ = """Convert DICOM dirs based on heuristic info - HeuDiConv
uses the dcmstack package and dcm2niix tool to convert DICOM directories or
tarballs into collections of NIfTI files following pre-defined heuristic(s)."""

REQUIRES = [
    'nibabel',
    'pydicom',
    'nipype',
    'pathlib',
]

TESTS_REQUIRES = [
    'six',
    'pytest',
    'mock',
    'tinydb',
    'inotify',
]

EXTRA_REQUIRES = {
    'tests': TESTS_REQUIRES,
    'extras': [],  # Requires patched version ATM ['dcmstack'],
    'datalad': ['datalad']
}

# Flatten the lists
EXTRA_REQUIRES['all'] = sum(EXTRA_REQUIRES.values(), [])



#
# Tests for reproin.py
#
from collections import OrderedDict
from .reproin import get_dups_marked


def test_get_dups_marked():
    no_dups = {('some',): [1]}
    assert get_dups_marked(no_dups) == no_dups

    assert get_dups_marked(
        OrderedDict([
            (('bu', 'du'), [1, 2]),
            (('smth',), [3]),
            (('smth2',), ['a', 'b', 'c'])
        ])) == \
        {
            ('bu__dup-01', 'du'): [1],
            ('bu', 'du'): [2],
            ('smth',): [3],
            ('smth2__dup-02',): ['a'],
            ('smth2__dup-03',): ['b'],
            ('smth2',): ['c']
        }




from heudiconv.cli.run import main as runner

import os
import os.path as op
from mock import patch
from six.moves import StringIO

from glob import glob
from os.path import join as pjoin, dirname
import csv
import re

import pytest
from .utils import TESTS_DATA_PATH

import logging
lgr = logging.getLogger(__name__)

try:
    from datalad.api import Dataset
except ImportError:  # pragma: no cover
    Dataset = None


# this will fail if not in project's root directory
def test_smoke_convertall(tmpdir):
    args = ("-c dcm2niix -o %s -b --datalad "
     "-s fmap_acq-3mm -d %s/{subject}/*"
     % (tmpdir, TESTS_DATA_PATH)
    ).split(' ')

    # complain if no heurisitic
    with pytest.raises(RuntimeError):
        runner(args)

    args.extend(['-f', 'convertall'])
    runner(args)


@pytest.mark.parametrize('heuristic', ['reproin', 'convertall'])
@pytest.mark.parametrize(
    'invocation', [
        "--files %s" % TESTS_DATA_PATH,    # our new way with automated groupping
        "-d %s/{subject}/* -s 01-fmap_acq-3mm" % TESTS_DATA_PATH # "old" way specifying subject
        # should produce the same results
    ])
@pytest.mark.skipif(Dataset is None, reason="no datalad")
def test_reproin_largely_smoke(tmpdir, heuristic, invocation):
    is_bids = True if heuristic == 'reproin' else False
    arg = "--random-seed 1 -f %s -c dcm2niix -o %s" \
          % (heuristic, tmpdir)
    if is_bids:
        arg += " -b"
    arg += " --datalad "
    args = (
        arg + invocation
    ).split(' ')

    # Test some safeguards
    if invocation == "--files %s" % TESTS_DATA_PATH:
        # Multiple subjects must not be specified -- only a single one could
        # be overridden from the command line
        with pytest.raises(ValueError):
            runner(args + ['--subjects', 'sub1', 'sub2'])

        if heuristic != 'reproin':
            # none other heuristic has mighty infotoids atm
            with pytest.raises(NotImplementedError):
                runner(args)
            return
    runner(args)
    ds = Dataset(str(tmpdir))
    assert ds.is_installed()
    assert not ds.repo.dirty
    head = ds.repo.get_hexsha()

    # and if we rerun -- should fail
    lgr.info(
        "RERUNNING, expecting to FAIL since the same everything "
        "and -c specified so we did conversion already"
    )
    with pytest.raises(RuntimeError):
        runner(args)

    # but there should be nothing new
    assert not ds.repo.dirty
    assert head == ds.repo.get_hexsha()

    # unless we pass 'overwrite' flag
    runner(args + ['--overwrite'])
    # but result should be exactly the same, so it still should be clean
    # and at the same commit
    assert ds.is_installed()
    assert not ds.repo.dirty
    assert head == ds.repo.get_hexsha()


@pytest.mark.parametrize(
    'invocation', [
        "--files %s" % TESTS_DATA_PATH,    # our new way with automated groupping
    ])
def test_scans_keys_reproin(tmpdir, invocation):
    args = "-f reproin -c dcm2niix -o %s -b " % (tmpdir)
    args += invocation
    runner(args.split())
    # for now check it exists
    scans_keys = glob(pjoin(tmpdir.strpath, '*/*/*/*/*/*.tsv'))
    assert(len(scans_keys) == 1)
    with open(scans_keys[0]) as f:
        reader = csv.reader(f, delimiter='\t')
        for i, row in enumerate(reader):
            if i == 0:
                assert(row == ['filename', 'acq_time', 'operator', 'randstr'])
            assert(len(row) == 4)
            if i != 0:
                assert(os.path.exists(pjoin(dirname(scans_keys[0]), row[0])))
                assert(re.match(
                    '^[\d]{4}-[\d]{2}-[\d]{2}T[\d]{2}:[\d]{2}:[\d]{2}$',
                    row[1]))


@patch('sys.stdout', new_callable=StringIO)
def test_ls(stdout):
    args = (
            "-f reproin --command ls --files %s"
            % (TESTS_DATA_PATH)
    ).split(' ')
    runner(args)
    out = stdout.getvalue()
    assert 'StudySessionInfo(locator=' in out
    assert 'Halchenko/Yarik/950_bids_test4' in out


def test_scout_conversion(tmpdir):
    tmppath = tmpdir.strpath
    args = (
        "-b -f reproin --files %s"
        % (TESTS_DATA_PATH)
    ).split(' ') + ['-o', tmppath]
    runner(args)

    assert not op.exists(pjoin(
        tmppath,
        'Halchenko/Yarik/950_bids_test4/sub-phantom1sid1/ses-localizer/anat'))

    assert op.exists(pjoin(
        tmppath,
        'Halchenko/Yarik/950_bids_test4/sourcedata/sub-phantom1sid1/'
        'ses-localizer/anat/sub-phantom1sid1_ses-localizer_scout.dicom.tgz'
    )
    )



import logging
import os
import os.path as op
from glob import glob
import re

from collections import defaultdict, OrderedDict

import tarfile
from tempfile import mkdtemp

from .dicoms import group_dicoms_into_seqinfos
from .utils import (
    docstring_parameter,
    StudySessionInfo,
    load_json,
    save_json,
    create_file_if_missing,
    json_dumps_pretty,
)

lgr = logging.getLogger(__name__)

_VCS_REGEX = '%s\.(?:git|gitattributes|svn|bzr|hg)(?:%s|$)' % (op.sep, op.sep)

@docstring_parameter(_VCS_REGEX)
def find_files(regex, topdir=op.curdir, exclude=None,
               exclude_vcs=True, dirs=False):
    """Generator to find files matching regex
    Parameters
    ----------
    regex: basestring
    exclude: basestring, optional
      Matches to exclude
    exclude_vcs:
      If True, excludes commonly known VCS subdirectories.  If string, used
      as regex to exclude those files (regex: `{}`)
    topdir: basestring, optional
      Directory where to search
    dirs: bool, optional
      Either to match directories as well as files
    """

    for dirpath, dirnames, filenames in os.walk(topdir):
        names = (dirnames + filenames) if dirs else filenames
        paths = (op.join(dirpath, name) for name in names)
        for path in filter(re.compile(regex).search, paths):
            path = path.rstrip(op.sep)
            if exclude and re.search(exclude, path):
                continue
            if exclude_vcs and re.search(_VCS_REGEX, path):
                continue
            yield path


def get_extracted_dicoms(fl):
    """Given a list of files, possibly extract some from tarballs
    For 'classical' heudiconv, if multiple tarballs are provided, they correspond
    to different sessions, so here we would group into sessions and return
    pairs  `sessionid`, `files`  with `sessionid` being None if no "sessions"
    detected for that file or there was just a single tarball in the list
    """
    # TODO: bring check back?
    # if any(not tarfile.is_tarfile(i) for i in fl):
    #     raise ValueError("some but not all input files are tar files")

    # tarfiles already know what they contain, and often the filenames
    # are unique, or at least in a unqiue subdir per session
    # strategy: extract everything in a temp dir and assemble a list
    # of all files in all tarballs

    # cannot use TempDirs since will trigger cleanup with __del__
    tmpdir = mkdtemp(prefix='heudiconvDCM')

    sessions = defaultdict(list)
    session = 0
    if not isinstance(fl, (list, tuple)):
        fl = list(fl)

    # needs sorting to keep the generated "session" label deterministic
    for i, t in enumerate(sorted(fl)):
        # "classical" heudiconv has that heuristic to handle multiple
        # tarballs as providing different sessions per each tarball
        if not tarfile.is_tarfile(t):
            sessions[None].append(t)
            continue

        tf = tarfile.open(t)
        # check content and sanitize permission bits
        tmembers = tf.getmembers()
        for tm in tmembers:
            tm.mode = 0o700
        # get all files, assemble full path in tmp dir
        tf_content = [m.name for m in tmembers if m.isfile()]
        # store full paths to each file, so we don't need to drag along
        # tmpdir as some basedir
        sessions[session] = [op.join(tmpdir, f) for f in tf_content]
        session += 1
        # extract into tmp dir
        tf.extractall(path=tmpdir, members=tmembers)

    if session == 1:
        # we had only 1 session, so no really multiple sessions according
        # to classical 'heudiconv' assumptions, thus just move them all into
        # None
        sessions[None] += sessions.pop(0)

    return sessions.items()


def get_study_sessions(dicom_dir_template, files_opt, heuristic, outdir,
                       session, sids, grouping='studyUID'):
    """Given options from cmdline sort files or dicom seqinfos into
    study_sessions which put together files for a single session of a subject
    in a study
    Two major possible workflows:
    - if dicom_dir_template provided -- doesn't pre-load DICOMs and just
      loads files pointed by each subject and possibly sessions as corresponding
      to different tarballs
    - if files_opt is provided, sorts all DICOMs it can find under those paths
    """
    study_sessions = {}
    if dicom_dir_template:
        dicom_dir_template = op.abspath(dicom_dir_template)

        # MG - should be caught by earlier checks
        # assert not files_opt  # see above TODO
        # assert sids
        # expand the input template

        if '{subject}' not in dicom_dir_template:
            raise ValueError(
                "dicom dir template must have {subject} as a placeholder for a "
                "subject id.  Got %r" % dicom_dir_template)
        for sid in sids:
            sdir = dicom_dir_template.format(subject=sid, session=session)
            files = sorted(glob(sdir))
            for session_, files_ in get_extracted_dicoms(files):
                if session_ is not None and session:
                    lgr.warning(
                        "We had session specified (%s) but while analyzing "
                        "files got a new value %r (using it instead)"
                        % (session, session_))
                # in this setup we do not care about tracking "studies" so
                # locator would be the same None
                study_sessions[StudySessionInfo(None,
                        session_ if session_ is not None else session,
                        sid)] = files_
    else:
        # MG - should be caught on initial run
        # YOH - what if it is the initial run?
        # prep files
        # assert files_opt
        files = []
        for f in files_opt:
            if op.isdir(f):
                files += sorted(find_files(
                    '.*', topdir=f, exclude_vcs=True, exclude="/\.datalad/"))
            else:
                files.append(f)

        # in this scenario we don't care about sessions obtained this way
        files_ = []
        for _, files_ex in get_extracted_dicoms(files):
            files_ += files_ex

        # sort all DICOMS using heuristic
        # TODO:  this one is not grouping by StudyUID but may be we should!
        seqinfo_dict = group_dicoms_into_seqinfos(files_,
            file_filter=getattr(heuristic, 'filter_files', None),
            dcmfilter=getattr(heuristic, 'filter_dicom', None),
            grouping=grouping)

        if not getattr(heuristic, 'infotoids', None):
            raise NotImplementedError(
                "For now, if no subj template is provided, requiring "
                "heuristic to have infotoids")

        if sids:
            if not (len(sids) == 1 and len(seqinfo_dict) == 1):
                raise RuntimeError(
                    "We were provided some subjects (%s) but "
                    "we can deal only "
                    "with overriding only 1 subject id. Got %d subjects and "
                    "found %d sequences" % (sids, len(sids), len(seqinfo_dict))
                )
            sid = sids[0]
        else:
            sid = None

        for studyUID, seqinfo in seqinfo_dict.items():
            # so we have a single study, we need to figure out its
            # locator, session, subject
            # TODO: Try except to ignore those we can't handle?
            # actually probably there should be a dedicated exception for
            # heuristics to throw if they detect that the study they are given
            # is not the one they would be willing to work on
            ids = heuristic.infotoids(seqinfo.keys(), outdir=outdir)
            # TODO:  probably infotoids is doomed to do more and possibly
            # split into multiple sessions!!!! but then it should be provided
            # full seqinfo with files which it would place into multiple groups
            lgr.info("Study session for %s" % str(ids))
            study_session_info = StudySessionInfo(
                ids.get('locator'),
                ids.get('session', session) or session,
                sid or ids.get('subject', None)
            )
            if study_session_info in study_sessions:
                #raise ValueError(
                lgr.warning(
                    "We already have a study session with the same value %s"
                    % repr(study_session_info))
                continue # skip for now
            study_sessions[study_session_info] = seqinfo
    return study_sessions



import os

def create_key(template, outtype=('nii.gz','dicom'), annotation_classes=None):
    if template is None or not template:
        raise ValueError('Template must be a valid format string')
    return (template, outtype, annotation_classes)


def infotodict(seqinfo):
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """
    t1 = create_key('anat/sub-{subject}_T1w')
    t2 = create_key('anat/sub-{subject}_T2w')
    rest = create_key('func/sub-{subject}_dir-{acq}_task-rest_run-{item:02d}_bold')
    face = create_key('func/sub-{subject}_task-face_run-{item:02d}_acq-{acq}_bold')
    gamble = create_key('func/sub-{subject}_task-gambling_run-{item:02d}_acq-{acq}_bold')
    conflict = create_key('func/sub-{subject}_task-conflict_run-{item:02d}_acq-{acq}_bold')
    dwi = create_key('dwi/sub-{subject}_dir-{acq}_run-{item:02d}_dwi')

    fmap_rest = create_key('fmap/sub-{subject}_acq-func{acq}_dir-{dir}_run-{item:02d}_epi')
    fmap_dwi = create_key('fmap/sub-{subject}_acq-dwi{acq}_dir-{dir}_run-{item:02d}_epi')

    info = {t1:[], t2:[], rest:[], face:[], gamble:[], conflict:[], dwi:[], fmap_rest:[], fmap_dwi:[]}

    for idx, s in enumerate(seqinfo):
        if (s.dim3 == 208) and (s.dim4 == 1) and ('T1w' in s.protocol_name):
            info[t1] = [s.series_id]
        if (s.dim3 == 208) and ('T2w' in s.protocol_name):
            info[t2] = [s.series_id]
        if (s.dim4 >= 99) and (('dMRI_dir98_AP' in s.protocol_name) or ('dMRI_dir99_AP' in s.protocol_name)):
            acq = s.protocol_name.split('dMRI_')[1].split('_')[0] + 'AP'
            info[dwi].append({'item': s.series_id, 'acq': acq})
        if (s.dim4 >= 99) and (('dMRI_dir98_PA' in s.protocol_name) or ('dMRI_dir99_PA' in s.protocol_name)):
            acq = s.protocol_name.split('dMRI_')[1].split('_')[0] + 'PA'
            info[dwi].append({'item': s.series_id, 'acq': acq})
        if (s.dim4 == 1) and (('dMRI_dir98_AP' in s.protocol_name) or ('dMRI_dir99_AP' in s.protocol_name)):
            acq = s.protocol_name.split('dMRI_')[1].split('_')[0]
            info[fmap_dwi].append({'item': s.series_id, 'dir': 'AP', 'acq': acq})
        if (s.dim4 == 1) and (('dMRI_dir98_PA' in s.protocol_name) or ('dMRI_dir99_PA' in s.protocol_name)):
            acq = s.protocol_name.split('dMRI_')[1].split('_')[0]
            info[fmap_dwi].append({'item': s.series_id, 'dir': 'PA', 'acq': acq})
        if (s.dim4 == 420) and ('rfMRI_REST_AP' in s.protocol_name):
            info[rest].append({'item': s.series_id, 'acq': 'AP'})
        if (s.dim4 == 420) and ('rfMRI_REST_PA' in s.protocol_name):
            info[rest].append({'item': s.series_id, 'acq': 'PA'})
        if (s.dim4 == 1) and ('rfMRI_REST_AP' in s.protocol_name):
            if seqinfo[idx + 1][9] != 420:
                continue
            info[fmap_rest].append({'item': s.series_id, 'dir': 'AP', 'acq': ''})
        if (s.dim4 == 1) and ('rfMRI_REST_PA' in s.protocol_name):
            info[fmap_rest].append({'item': s.series_id, 'dir': 'PA', 'acq': ''})
        if (s.dim4 == 346) and ('tfMRI_faceMatching_AP' in s.protocol_name):
            info[face].append({'item': s.series_id, 'acq': 'AP'})
        if (s.dim4 == 346) and ('tfMRI_faceMatching_PA' in s.protocol_name):
            info[face].append({'item': s.series_id, 'acq': 'PA'})
        if (s.dim4 == 288) and ('tfMRI_conflict_AP' in s.protocol_name):
            info[conflict].append({'item': s.series_id, 'acq': 'AP'})
        if (s.dim4 == 288) and ('tfMRI_conflict_PA' in s.protocol_name):
            info[conflict].append({'item': s.series_id, 'acq': 'PA'})
        if (s.dim4 == 223) and ('tfMRI_gambling_AP' in (s.protocol_name)):
            info[gamble].append({'item': s.series_id, 'acq': 'AP'})
        if (s.dim4 == 223) and ('tfMRI_gambling_PA' in s.protocol_name):
            info[gamble].append({'item': s.series_id, 'acq': 'PA'})
    return info



import os


def create_key(template, outtype=('nii.gz',), annotation_classes=None):
    if template is None or not template:
        raise ValueError('Template must be a valid format string')
    return template, outtype, annotation_classes


def infotodict(seqinfo):
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    """

    data = create_key('run{item:03d}')
    info = {data: []}
    last_run = len(seqinfo)

    for s in seqinfo:
        """
        The namedtuple `s` contains the following fields:

        * total_files_till_now
        * example_dcm_file
        * series_id
        * dcm_dir_name
        * unspecified2
        * unspecified3
        * dim1
        * dim2
        * dim3
        * dim4
        * TR
        * TE
        * protocol_name
        * is_motion_corrected
        * is_derived
        * patient_id
        * study_description
        * referring_physician_name
        * series_description
        * image_type
        """

        info[data].append(s.series_id)
    return info
